package vpn

import (
	"crypto/sha256"
	"fmt"
	tun "git.sydch.com/Go-Module/atom-device"
	"git.sydch.com/Go-Module/atom/utils"
	"git.sydch.com/Go-Module/atom/vpn/ping"
	"github.com/golang/snappy"
	"github.com/xtaci/kcp-go"
	"golang.org/x/crypto/pbkdf2"
	"time"
)

type (
	Client struct {
		tun     *tun.Device
		session *kcp.UDPSession
		status  bool

		key    []byte
		config *Config

		inStream  chan *tun.Packet
		outStream chan *tun.Packet
	}
)

func NewClient(config *Config) *Client {
	c := new(Client)
	c.key = pbkdf2.Key([]byte(config.Password), []byte(KcpKey), 4096, 32, sha256.New)[:16]
	c.config = config
	c.status = false
	utils.SetKey(DataKey)

	c.inStream = make(chan *tun.Packet, 200)
	c.outStream = make(chan *tun.Packet, 200)

	return c
}

func (c *Client) StartTun() error {
	tunDevice, err := tun.NewDevice(nil, "Atom", c.config.IFaceIp, c.config.IFaceNet)
	if err != nil {
		return fmt.Errorf("Tunnel网卡初始化失败\n %v\n", err.Error())
	}
	c.tun = tunDevice

	go c.ReadTun()
	go c.WriteTun()
	return nil
}

func (c *Client) ReadTun() {
	for packet := range c.tun.OutboundChan() {
		c.outStream <- packet
	}
}

func (c *Client) WriteTun() {
	for packet := range c.inStream {
		err := c.tun.WritePacket(packet, packet.Src)
		if err != nil {
			c.inStream <- packet
			fmt.Println("包写入错误")
		}
	}
}

func (c *Client) StartKCP() error {
	crypt, err := kcp.NewAESBlockCrypt(c.key[:16])
	if err != nil {
		return fmt.Errorf("生成证书错误: %v", err.Error())
	}

	if c.session, err = kcp.DialWithOptions(c.config.NetAddr, crypt, 10, 3); err != nil {
		return fmt.Errorf("KCP服务启动失败: %v", err.Error())
	}

	c.session.SetWindowSize(SndWnd, RcvWnd)
	c.session.SetACKNoDelay(true)
	c.session.SetStreamMode(false)

	if err := c.session.SetDSCP(DSCP); err != nil {
		return fmt.Errorf("DSCP设置失败, Val: %v\n%v", DSCP, err.Error())
	}
	if err := c.session.SetReadBuffer(SockBuf); err != nil {
		return fmt.Errorf("ReadBuffer设置失败, Val: %v\n%v", SockBuf, err.Error())
	}
	if err := c.session.SetWriteBuffer(SockBuf); err != nil {
		return fmt.Errorf("ReadBuffer设置失败, Val: %v\n%v", SockBuf, err.Error())
	}

	c.status = true

	return nil
}

func (c *Client) Start() {
	go c.tunToKcp()
	for {
		if !c.status {
			err := c.StartKCP()
			if err != nil {
				fmt.Println(err)
				continue
			}
			c.status = true
		}
		go c.CheckKCPSessionAlive()
		c.kcpToTun()
	}
}

func (c *Client) CheckKCPSessionAlive() {
	for {
		time.Sleep(time.Second * 5)
		if !ping.Ping(c.config.ServerIFaceIP) {
			c.status = false
			c.session.Close()
			fmt.Println("链路检测失败, 重连中.....")
			break
		}
	}
}

func (c *Client) kcpToTun() {
	defer c.session.Close()

	tempData := make([]byte, tun.InterfaceMTU*2)
	//var tempData []byte
	for c.status {
		read, err := c.session.Read(tempData)
		if err != nil {
			fmt.Println(err.Error())
			continue
		}

		data := tempData[:read]

		// 是否用压缩
		if c.config.Compress {
			data, err = snappy.Decode(nil, data)
			if err != nil {
				fmt.Println(err.Error())
				continue
			}
		}

		// 是否启动混淆
		if c.config.Obfs {
			data = utils.XOR(data)
		}

		packet := new(tun.Packet)
		packet.Packet = data
		packet.ParseBuffer()
		packet.Parse()
		c.inStream <- packet
	}
}

func (c *Client) tunToKcp() {
	var data []byte
	for packet := range c.outStream {
		if c.status {
			// 对网卡上来的数据进行深拷贝
			data = utils.DeepCopyBytes(packet.Packet)

			// 是否启动混淆
			if c.config.Obfs {
				data = utils.XOR(data)
			}

			// 是否用压缩
			if c.config.Compress {
				data = snappy.Encode(nil, data)
			}

			_, err := c.session.Write(data)
			if err != nil {
				fmt.Println(err.Error())
				c.outStream <- packet
				continue
			}
		} else {
			c.outStream <- packet
		}
	}
}
