package vpn

import (
	"crypto/sha256"
	"fmt"
	tun "git.sydch.com/Go-Module/atom-device"
	"git.sydch.com/Go-Module/atom/utils"
	"git.sydch.com/Go-Module/atom/vpn/ping"
	"github.com/golang/snappy"
	"github.com/xtaci/kcp-go"
	"golang.org/x/crypto/pbkdf2"
	"sync"
	"time"
)

type (
	Server struct {
		tun  *tun.Device
		conn *kcp.Listener

		Cache  *sync.Map
		key    []byte
		config *Config

		inBoundCh  chan *tun.Packet // inBoundCh 服务端入栈队列
		outBoundCh chan *tun.Packet // outBoundCh 服务端出栈队列
	}

	ClientSession struct {
		localIP string          // localIP 客户端IP地址
		session *kcp.UDPSession // session 客户端会话
		status  bool            // status 客户端在线状态 用于控制客户端后台的任务

		outBoundCh chan *tun.Packet // outBoundCh 客户端出栈队列
		inBoundCh  chan *tun.Packet // inBoundCh 客户端入栈队列
	}
)

func NewServer(config *Config) *Server {
	s := new(Server)
	s.key = pbkdf2.Key([]byte(config.Password), []byte(KcpKey), 4096, 32, sha256.New)[:16]
	s.Cache = new(sync.Map)
	s.config = config
	s.inBoundCh = make(chan *tun.Packet, 200)
	s.outBoundCh = make(chan *tun.Packet, 200)
	utils.SetKey(DataKey)
	return s
}

func (s *Server) StartTun() error {
	tunDevice, err := tun.NewDevice(nil, "Atom", s.config.IFaceIp, s.config.IFaceNet)
	if err != nil {
		return fmt.Errorf("Tunnel网卡初始化失败\n %v\n", err.Error())
	}
	s.tun = tunDevice

	go s.ReadTun()
	go s.WriteTun()
	return nil
}

func (s *Server) StartKcp() error {
	crypt, err := kcp.NewAESBlockCrypt(s.key)
	if err != nil {
		return fmt.Errorf("生成证书错误: %v", err.Error())
	}

	if s.conn, err = kcp.ListenWithOptions(s.config.NetAddr, crypt, 10, 3); err != nil {
		return fmt.Errorf("KCP服务启动失败: %v", err.Error())
	}

	if err := s.conn.SetDSCP(DSCP); err != nil {
		return fmt.Errorf("DSCP设置失败, Val: %v\n%v", DSCP, err.Error())
	}
	if err := s.conn.SetReadBuffer(SockBuf); err != nil {
		return fmt.Errorf("ReadBuffer设置失败, Val: %v\n%v", SockBuf, err.Error())
	}
	if err := s.conn.SetWriteBuffer(SockBuf); err != nil {
		return fmt.Errorf("ReadBuffer设置失败, Val: %v\n%v", SockBuf, err.Error())
	}

	return nil
}

func (s *Server) Start() {
	go s.serverToClient()
	go func() {
		for {
			s.Cache.Range(func(key, value any) bool {

				if !ping.Ping(key.(string)) {
					fmt.Printf("客户端IP: %v, 地址: %v \n", key, &value)
					s.Cache.Delete(key.(string))
					client := value.(*ClientSession)
					s.CloseClient(client, "close1")
				}
				return true
			})
			time.Sleep(time.Second * 1)
		}
	}()
	for {
		session, err := s.conn.AcceptKCP()
		if err != nil {
			fmt.Printf("Session 链接失败 %v", err.Error())
			continue
		}
		fmt.Println("在处理新链接")
		session.SetWindowSize(SndWnd, RcvWnd)
		session.SetACKNoDelay(true)
		session.SetStreamMode(false)

		// TODO - 为客户端提供进出栈的功能

		// 初始化客户端通道栈
		c := new(ClientSession)
		c.inBoundCh = make(chan *tun.Packet, 200)
		c.outBoundCh = make(chan *tun.Packet, 200)
		c.session = session
		c.status = true

		// TODO - 启动向客户端发生信息的线程
		go s.clientIn(c)
		go s.clintOut(c)

	}

}

// toClient 处理客户端上来的数据包 进行分类然后分发
func (s *Server) clientIn(c *ClientSession) {
	fmt.Printf("协程<1>被开启 IP: %v \n", c.session.RemoteAddr().String())
	tempData := make([]byte, tun.InterfaceMTU)
	for {
		// 从客户端读取数据包头 来确定是否为正确包
		read, err := c.session.Read(tempData)
		if err != nil {
			fmt.Println(err.Error())
			break
		}

		data := tempData[:read]

		// 是否启用压缩
		if s.config.Compress {
			data, err = snappy.Decode(nil, data)
			if err != nil {
				fmt.Println(err.Error())
				continue
			}
		}

		// 是否用启用混淆
		if s.config.Obfs {
			data = utils.XOR(data)
		}

		// TODO - 将全部包解析完毕 并分发给指定客户端的地址
		packet := s.tun.GetTempPacket()
		packet.Packet = data
		packet.ParseBuffer()
		packet.Parse()

		c.localIP = packet.Src.String()

		if obj, ok := s.Cache.Load(c.localIP); ok {
			client := obj.(*ClientSession)
			if client.session != c.session {
				s.CloseClient(client, "close2")
			}
		}

		//packet := new(tun.Packet)
		//packet.Packet = data
		//packet.ParseBuffer()
		//packet.Parse()

		s.Cache.Store(c.localIP, c)

		// 如果是组播地址则放入全部客户端队列
		if packet.Dst.IsMulticast() || packet.Dst.String() == "255.255.255.255" {
			if packet.Src.String() != s.config.IFaceIp.String() {
				tempPacket := s.tun.GetTempPacket()
				*tempPacket = *packet
				SafeSend(s.inBoundCh, tempPacket)
			}

			s.Cache.Range(func(key, value any) bool {
				ip := key.(string)

				if ip == packet.Src.String() {
					return true
				}
				client := value.(*ClientSession)
				tempPacket := s.tun.GetTempPacket()
				*tempPacket = *packet
				SafeSend(client.outBoundCh, tempPacket)

				return true
			})

			s.tun.PutTempPacket(packet)
			continue
		}

		// 如果是服务端地址则放入服务端处理队列
		if packet.Dst.String() == s.config.IFaceIp.String() {
			tempPacket := s.tun.GetTempPacket()
			*tempPacket = *packet
			SafeSend(s.inBoundCh, tempPacket)
			s.tun.PutTempPacket(packet)
			continue
		}

		// 如果是其他客户端的包则写入
		if obj, ok := s.Cache.Load(packet.Dst.String()); ok {
			client := obj.(*ClientSession)
			tempPacket := s.tun.GetTempPacket()
			*tempPacket = *packet
			SafeSend(client.outBoundCh, tempPacket)
			s.tun.PutTempPacket(packet)
			continue
		}
	}

	fmt.Println("协程<1>被关闭")
}

func (s *Server) clintOut(c *ClientSession) {
	fmt.Printf("协程<2>被开启 IP: %v \n", c.session.RemoteAddr().String())
	var data []byte
	for {
		packet, open := <-c.outBoundCh
		if !open {
			break
		}
		// 对网卡上来的数据进行深拷贝
		data = utils.DeepCopyBytes(packet.Packet)

		s.tun.PutTempPacket(packet)

		// 是否启动混淆
		if s.config.Obfs {
			data = utils.XOR(data)
		}

		// 是否用压缩
		if s.config.Compress {
			data = snappy.Encode(nil, data)
		}

		_, err := c.session.Write(data)
		if err != nil {
			fmt.Println(err.Error())
			break
		}

	}
	fmt.Println("协程<2>被关闭")

}

func (s *Server) serverToClient() {
	var data []byte
	for packet := range s.outBoundCh {
		if val, ok := s.Cache.Load(packet.Dst.String()); ok {

			// 对网卡上来的数据进行深拷贝
			data = utils.DeepCopyBytes(packet.Packet)

			s.tun.PutTempPacket(packet)

			// 是否启动混淆
			if s.config.Obfs {
				data = utils.XOR(data)
			}

			// 是否用压缩
			if s.config.Compress {
				data = snappy.Encode(nil, data)
			}

			client := val.(*ClientSession)
			_, err := client.session.Write(data)
			if err != nil {
				fmt.Println(err.Error())
				continue
			}
		}
	}
}

func (s *Server) CloseClient(c *ClientSession, info string) {
	fmt.Println("我被执行了: ", info)
	c.status = false
	s.Cache.Delete(c.localIP)
	err := c.session.Close()
	if err != nil {
		fmt.Println(err.Error())
	}

	// 等待全部协程关闭
	SafeClose(c.inBoundCh)
	SafeClose(c.outBoundCh)
	for packet := range c.inBoundCh {
		s.tun.PutTempPacket(packet)
	}
	for packet := range c.outBoundCh {
		s.tun.PutTempPacket(packet)
	}

}

func (s *Server) ReadTun() {
	for packet := range s.tun.OutboundChan() {
		s.outBoundCh <- packet
	}
}

func (s *Server) WriteTun() {
	for packet := range s.inBoundCh {
		err := s.tun.WritePacket(packet, packet.Src)
		if err != nil {
			s.tun.PutTempPacket(packet)
			fmt.Println("包写入错误")
		}
	}
}

func SafeSend(ch chan *tun.Packet, value *tun.Packet) (closed bool) {
	defer func() {
		if recover() != nil {
			// the return result can be altered
			// in a defer function call
			closed = true
		}
	}()

	ch <- value  // panic if ch is closed
	return false // <=> closed = false; return
}

func SafeClose(ch chan *tun.Packet) (justClosed bool) {
	defer func() {
		if recover() != nil {
			fmt.Println("通道已经被关闭")

			justClosed = false
		}
	}()

	// assume ch != nil here.
	close(ch) // panic if ch is closed
	return true
}
