package vpnDemo

import (
	"crypto/sha256"
	"fmt"
	tun "git.sydch.com/Go-Module/atom-device"
	"git.sydch.com/Go-Module/atom/utils"
	"github.com/golang/snappy"
	"github.com/google/uuid"
	"github.com/xtaci/kcp-go"
	"go.uber.org/zap"
	"golang.org/x/crypto/pbkdf2"
	"net"
	"sync"
)

type (
	Server struct { // Server 服务端缓存类
		tun    *tun.Device   // tun 虚拟网卡
		conn   *kcp.Listener // conn Kcp服务端
		config *Config       // config 服务端配置文件

		netPeers sync.Map           // netPeers 网络在线用户 使用ip作为索引
		logger   *zap.SugaredLogger // logger 日志系统

		allPeers sync.Map // allPeers 新建的客户端 以uuid为索引 当链接成功后删除

		inBoundCh  chan *tun.Packet // inBoundCh 服务端入栈队列
		outBoundCh chan *tun.Packet // outBoundCh 服务端出栈队列
	}
	VpnPeer struct {
		localIP net.IP          // localIP 客户端本地ip地址
		session *kcp.UDPSession // session 客户端KCP链接句柄
		uuid    uuid.UUID       // uuid 客户端回话的uuid

		inboundCh  chan *tun.Packet // inboundCh 客户端入栈队列 外部 -> 客户端Session
		outboundCh chan *tun.Packet // outboundCh 客户端出栈队列 客户端Session -> 外部
		wg         sync.WaitGroup   // wg 管理协程的线程组 在执行关闭方法时应同步被关闭
		//status     bool             // status 客户端状态 用于控制循环状态(寻找更好地替代方案)
	}
)

func NewServer(config *Config) *Server {
	s := new(Server)
	s.config = config
	s.inBoundCh = make(chan *tun.Packet, 200)
	s.outBoundCh = make(chan *tun.Packet, 200)
	utils.SetKey(DataKey)
	return s
}

func (s *Server) StartTun() error {
	tunDevice, err := tun.NewDevice(nil, "Atom", s.config.IFaceIp, s.config.IFaceNet)
	if err != nil {
		return fmt.Errorf("Tunnel网卡初始化失败\n %v\n", err.Error())
	}
	s.tun = tunDevice

	go s.readTun()
	go s.writeTun()
	return nil
}

func (s *Server) StartKcp() error {
	key := pbkdf2.Key([]byte(s.config.Password), []byte(KcpKey), 4096, 32, sha256.New)[:16]
	crypt, err := kcp.NewAESBlockCrypt(key[:16])
	if err != nil {
		return fmt.Errorf("生成证书错误: %v", err.Error())
	}

	if s.conn, err = kcp.ListenWithOptions(s.config.NetAddr, crypt, 10, 3); err != nil {
		return fmt.Errorf("KCP服务启动失败: %v", err.Error())
	}

	if err := s.conn.SetDSCP(DSCP); err != nil {
		return fmt.Errorf("DSCP设置失败, Val: %v\n%v", DSCP, err.Error())
	}
	if err := s.conn.SetReadBuffer(SockBuf); err != nil {
		return fmt.Errorf("ReadBuffer设置失败, Val: %v\n%v", SockBuf, err.Error())
	}
	if err := s.conn.SetWriteBuffer(SockBuf); err != nil {
		return fmt.Errorf("ReadBuffer设置失败, Val: %v\n%v", SockBuf, err.Error())
	}
	return nil
}

func (s *Server) Start() {
	go s.serverToClient()
	for {
		session, err := s.conn.AcceptKCP()
		if err != nil {
			fmt.Printf("Client Conn Err: %v\n", err.Error())
		}
		fmt.Printf("新客户端连入 IP: %v\n", session.RemoteAddr().String())
		session.SetWindowSize(SndWnd, RcvWnd)
		session.SetACKNoDelay(true)
		session.SetStreamMode(false)

		c := new(VpnPeer)
		c.inboundCh = make(chan *tun.Packet, 200)
		c.outboundCh = make(chan *tun.Packet, 200)
		c.session = session
		c.uuid = uuid.New()

		s.allPeers.Store(c.uuid, c)

		c.wg.Add(2)
		go s.backgroundInboundHandler(c)
		go s.backgroundOutboundHandler(c)
	}
}

//func (s *Server) closeErrorClient() {
//	s.allPeers.Range(func(key, value any) bool {
//
//	})
//}

// serverToClient 服务端接受到的数据发送到指定客户端 TODO - 等待客户端通讯完成后制作
func (s *Server) serverToClient() {
	for packet := range s.outBoundCh {
		// 实体类读加锁 避免多个对象同时操作

		// 根据IP或者客户端的实体类方便将数据放入指定客户端的出栈队列中
		obj, ok := s.netPeers.Load(packet.Dst.String())
		if !ok {
			// 如果无这个客户端则把这个包扔回缓冲区
			s.tun.PutTempPacket(packet)
			continue
		}
		peer := obj.(*VpnPeer)

		// 这里更改为并不直接向客户端发送包 避免出现多个协程同时操作发送的问题 减少内存逃逸的可能
		select {
		case peer.outboundCh <- packet:
		default:
			s.tun.PutTempPacket(packet)
		}
	}
}

// readTun 从Tun网卡出栈队列读取到的包缓存到出栈缓冲队列
func (s *Server) readTun() {
	for packet := range s.tun.OutboundChan() {
		select {
		case s.outBoundCh <- packet:
		default:
			s.tun.PutTempPacket(packet)
		}
	}
}

// writeTun 从服务端入栈队列拉取包后写入到网卡
func (s *Server) writeTun() {
	for packet := range s.inBoundCh {
		err := s.tun.WritePacket(packet, packet.Src)
		if err != nil {
			s.tun.PutTempPacket(packet)
			fmt.Printf("Tun Err: 网卡通道写入错误 %v\n", err.Error())
		}
	}
}

// backgroundInboundHandler 集中收集客户端信息 收集客户端上来的数据包 将数据分发到对应的位置
func (s *Server) backgroundInboundHandler(peer *VpnPeer) {
	defer peer.wg.Done()
	for {
		packet := s.tun.GetTempPacket()
		packet.Packet = make([]byte, tun.InterfaceMTU*2)
		read, err := peer.session.Read(packet.Packet)
		if err != nil {
			fmt.Printf("KCP Session Err: 无法正确的读取Session : %v\n", err.Error())
			s.tun.PutTempPacket(packet)
			break
		}

		packet.Packet = packet.Packet[:read]

		// TODO - 是否启用压缩 (Google压缩算法似乎有点问题 - 回导致解压失败 目前在寻找解决方案)

		// 是否启用压缩
		if s.config.Compress {
			packet.Packet, err = snappy.Decode(nil, packet.Packet)
			if err != nil {
				fmt.Println(err.Error())
				continue
			}
		}

		// 是否用启用混淆
		if s.config.Obfs {
			packet.Packet = utils.XOR(packet.Packet)
		}

		// 将数据转为对应的以太网帧
		packet.ParseBuffer()
		// 获取数据包内的Dst IP及Src IP
		packet.Parse()

		if packet.IsIPv6 {
			continue
		}

		// 每次收到数据包后校准Src
		peer.localIP = packet.Src
		// 收到后判断过去是否链接过 如果链接过则清空过期的Session TODO - 可能为一处内存逃逸的可能位置
		if obj, ok := s.netPeers.Load(peer.localIP.String()); ok {
			oldPeer := obj.(*VpnPeer)
			if oldPeer.session != peer.session {
				fmt.Printf("我被执行了\n")
				oldPeer.Close(s)
			}
		}

		s.allPeers.Delete(peer.uuid)
		// 将现在客户端的指针地址存入Map - 注意加解锁操作
		s.netPeers.Store(peer.localIP.String(), peer)

		// 开始针对包开始处理流程
		// TODO - 处理广播的以太网帧

		// 如果包的地址是服务端地址则扔入服务端处理队列
		if packet.Dst.Equal(s.config.IFaceIp) {
			select {
			case s.inBoundCh <- packet:
			default:
				s.tun.PutTempPacket(packet)
				continue
			}

		}

		// 如果是其他客户端的包则写入其他客户端
		if obj, ok := s.netPeers.Load(packet.Dst.String()); ok {
			sendPeer := obj.(*VpnPeer)
			select {
			case sendPeer.outboundCh <- packet:
			default:
				s.tun.PutTempPacket(packet)
				continue
			}
		}
	}
	fmt.Println("线程1关闭")
}

// backgroundInboundHandler 集中向客户端写入信息 将出栈信息写入指定客户端的session
func (s *Server) backgroundOutboundHandler(peer *VpnPeer) {
	defer peer.wg.Done()
	fmt.Printf("协程<2>被开启 IP: %v \n", peer.session.RemoteAddr().String())
	for {
		packet, open := <-peer.outboundCh
		if !open {
			break
		}

		// 是否启用混淆
		if s.config.Obfs {
			packet.Packet = utils.XOR(packet.Packet)
		}
		// TODO - 是否启用压缩 (Google压缩算法似乎有点问题 - 回导致解压失败 目前在寻找解决方案)
		// 是否用压缩
		if s.config.Compress {
			packet.Packet = snappy.Encode(nil, packet.Packet)
		}

		_, err := peer.session.Write(packet.Packet)
		if err != nil {
			fmt.Printf("KCP Session Err: 无法正确的写入Session: %v\n", err.Error())
			break
		}
		s.tun.PutTempPacket(packet)
	}
	fmt.Println("线程2关闭")

}

func (v *VpnPeer) Close(s *Server) {

	// 关闭Session 避免更多数据被加入
	err := v.session.Close()
	if err != nil {
		fmt.Printf("Close Client Session Err: %v\n", err)
	}

	s.netPeers.Delete(v.localIP.String())

	// 关闭通道 - 需要注意是否回出现可能的错误
	SafeClose(v.inboundCh)
	SafeClose(v.outboundCh)

	v.wg.Wait()

	for packet := range v.inboundCh {
		s.tun.PutTempPacket(packet)
	}
	for packet := range v.inboundCh {
		s.tun.PutTempPacket(packet)
	}

}

func SafeSend(ch chan *tun.Packet, value *tun.Packet) (closed bool) {
	defer func() {
		if recover() != nil {
			// the return result can be altered
			// in a defer function call
			closed = true
		}
	}()

	ch <- value  // panic if ch is closed
	return false // <=> closed = false; return
}

func SafeClose(ch chan *tun.Packet) (justClosed bool) {
	defer func() {
		if recover() != nil {
			fmt.Println("通道已经被关闭")

			justClosed = false
		}
	}()

	// assume ch != nil here.
	close(ch) // panic if ch is closed
	return true
}
